#!/usr/bin/perl

package Cognition::Export::Calendar;
use Cognition::Export;
use Cognition::Export::CalComponent;
use HTML::Entities qw(encode_entities_numeric);;

sub is
{
	my $subject = shift;
	return unless ($subject);
	
	return 1 if (Cognition::Export::Feed::is($subject));
	
	return
		unless (defined $subject->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'));
	foreach my $Type ( @{$subject
		->property('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
		->valObjects} )
	{
		my $type = $Type->toString;
		return 1 if ($type eq 'urn:ietf:rfc:2445#vcalendar');
	}
	
	return;
}

sub outputs
{
	return [
		{
			code  => 'ICS',
			name  => 'iCalendar',
			type  => 'text/directory',
			multi => 1
		}
	];
}

sub outICS
{
	my $triples = shift;
	my $subject = shift;
	my $S = findSubject($triples, $subject);
	
	# Note that the W3C RDF iCal vocab (which Cognition does support) doesn't
	# actually support the concept of a "calendar" as such: just events, todos,
	# journal entries, etc. If you want to export standalone components like
	# that, then use Cognition::Export::CalComponent::multiICS. This function
	# only supports actual calendar objects (feeds are a special case of that).
	# A calendar object can be created in RDF my giving it an rdf:type of
	# "urn:ietf:rfc:2445#vcalendar" and linking to component resources using
	# either "urn:ietf:rfc:2445#vevent" and its bretheren, or dcterms:hasPart.
	
	my @components = allOf($S,
		'urn:ietf:rfc:2445#vevent',
		'urn:ietf:rfc:2445#vtodo',
		'urn:ietf:rfc:2445#valarm',
		'urn:ietf:rfc:2445#vjournal',
		'urn:ietf:rfc:2445#vfreebusy',
		'http://purl.org/dc/terms/hasPart',
		'http://purl.org/rss/1.0/item',
		'urn:ietf:rfc:4287#entry'
	);
	
	my %qualified_components;
	foreach my $c (@components)
	{
		my $C = findSubject($triples, $c);
		next unless($C);
		
		if (Cognition::Export::CalComponent::is($C))
		{
			unless (defined $qualified_components{$c})
				{ $qualified_components{$c} = 1; }
			
			# Any alarm which is embedded within another component is disqualified.
			my @alarms = allOf($C, 'urn:ietf:rfc:2445#valarm');
			foreach my $a (@alarms)
			{
				$qualified_components{$a} = 0;
			}
		}
	}
	
	@components = ();
	foreach my $c (keys %qualified_components)
	{
		if ($qualified_components{$c})
		{
			push @components, $c;
		}
	}
	
	my $rv = "BEGIN:VCALENDAR\r\n";
	
	my $ver = firstOf($S, 'urn:ietf:rfc:2445#version');
	if ($ver)
		{ $rv .= "VERSION:$ver\r\n"; }
	else
		{ $rv .= "VERSION:2.0\r\n"; }

	my $prodid = firstOf($S, 'urn:ietf:rfc:2445#prodid');
	if ($prodid)
		{ $rv .= "PRODID:$prodid\r\n"; }
	else
		{ $rv .= "PRODID:+//IDN buzzword.org.uk//NONSGML Cognition 0.1//EN\r\n"; }

	my $method = uc(firstOf($S, 'urn:ietf:rfc:2445#method'));
	if ($method)
		{ $rv .= "METHOD:$method\r\n"; }
	else
		{ $rv .= "METHOD:PUBLISH\r\n"; }

	my $title = firstOf($S,
		'urn:ietf:rfc:2445#x-wr-calname',
		'urn:ietf:rfc:4287#title',
		'http://purl.org/rss/1.0/title',
		'http://purl.org/dc/terms/title',
		'http://purl.org/dc/elements/1.1/title');
	$rv .= "X-WR-CALNAME:$title\r\n" if ($title);

	my $desc = firstOf($S,
		'urn:ietf:rfc:2445#x-wr-caldesc',
		'http://purl.org/rss/1.0/description',
		'http://purl.org/dc/terms/description',
		'http://purl.org/dc/elements/1.1/description',
		'http://purl.org/dc/terms/abstract',
		'urn:ietf:rfc:4287#subtitle');
	$rv .= "X-WR-CALDESC:$desc\r\n" if ($desc);	
	
	my $scale = uc(firstOf($S, 'urn:ietf:rfc:2445#calscale'));
	if ($scale)
		{ $rv .= "CALSCALE:$calscale\r\n"; }
	else
		{ $rv .= "CALSCALE:GREGORIAN\r\n"; }

	foreach my $c (@components)
	{
		$rv .= Cognition::Export::CalComponent::_outICS($triples, $c);
	}

	$rv .= "END:VCALENDAR\r\n";
	
	return $rv;
}

sub multiICS
{
	my $triples = shift;
	my $rv = '';
	foreach my $s (keys %$triples)
	{
		$rv .= outICS($triples, $s)
			if (defined $triples->{$s} && is($triples->{$s}));
	}

	return $rv;
}

1;
